#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2024-11-27
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

import unittest
from pathlib import Path
from robomeshcat import Robot


class TestRobot(unittest.TestCase):
    def test_q_by_name(self):
        robot = Robot(urdf_path=Path(__file__).parent / 'test_urdf.urdf')

        robot["shoulder_pan_joint"] = 0.1
        self.assertAlmostEqual(robot["shoulder_pan_joint"], 0.1)
        self.assertAlmostEqual(robot[0], 0.1)

    def test_q_by_name_wrong_key(self):
        robot = Robot(urdf_path=Path(__file__).parent / 'test_urdf.urdf')
        self.assertRaises(KeyError, lambda: robot["wrong_key"])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-12-6
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#
from __future__ import annotations

import numpy as np
from meshcat import geometry as g

from . import Object


class Human(Object):
    def __init__(
        self,
        pose=None,
        color: list[float] | None = None,
        opacity: float = 1.0,
        name: str | None = None,
        use_vertex_colors: bool = False,
        show_wireframe: bool = False,
        **kwargs,
    ) -> None:
        from smplx import SMPLX  # we import SMPLX here on purpose, so that smplx dependencies are optional

        self.smplx_model = SMPLX(**kwargs)
        super().__init__(None, pose, color if not use_vertex_colors else [1, 1, 1], None, opacity, name)

        self._use_vertex_colors = use_vertex_colors
        self._show_wireframe = show_wireframe
        clr = self._color_from_input(color)[np.newaxis, :]
        self._vertex_colors = np.repeat(clr, self.smplx_model.get_num_verts(), 0) if use_vertex_colors else None
        self.update_vertices(set_object=False)  # this will create a geometry

        "Additional properties that are modifiable "
        self._morph_target_influences = None

    @property
    def _material(self):
        mat = super()._material
        mat.wireframe = self._show_wireframe
        mat.vertexColors = self._use_vertex_colors
        return mat

    def get_vertices(self, **kwargs):
        """Return vertices of the mesh for the given smplx parameters."""
        output = self.smplx_model(return_verts=True, **kwargs)
        return output.vertices.detach().cpu().numpy().squeeze()

    def update_vertices(self, vertices=None, vertices_colors=None, set_object=True):
        if self._is_animation():
            print(
                'Update vertices of the mesh will recreate the geometry. It cannot be used in animation for '
                'which you should use this.add_morph function'
            )
            return
        self._geometry = TriangularMeshGeometryWithMorphAttributes(
            vertices=np.asarray(vertices) if vertices is not None else self.get_vertices(),
            faces=self.smplx_model.faces,
            color=np.asarray(vertices_colors) if vertices_colors is not None else self._vertex_colors,
            morph_positions=[],
            morph_colors=[],
        )
        if set_object:
            self._set_object()

    def add_morph(self, vertices, vertex_colors=None):
        """Add new morphology through which we can create animations."""
        self._geometry.morph_positions.append(vertices)
        if vertex_colors is not None:
            self._geometry.morph_colors.append(np.asarray(vertex_colors))

    def display_morph(self, morph_id):
        """Set morphTargetInfluences to display only the given morph_id."""
        self._morph_target_influences = [0] * self._geometry.number_of_morphs()
        if morph_id is not None:
            self._morph_target_influences[morph_id] = 1

    def _set_morph_property(self):
        if self._morph_target_influences is None:
            self._morph_target_influences = [0] * len(self._geometry.morph_positions)
        self._set_property('morphTargetInfluences', self._morph_target_influences, 'vector')

    def _reset_all_properties(self):
        super()._reset_all_properties()
        self._set_morph_property()


class TriangularMeshGeometryWithMorphAttributes(g.TriangularMeshGeometry):
    def __init__(self, morph_positions=None, morph_colors=None, **kwargs):
        super(TriangularMeshGeometryWithMorphAttributes, self).__init__(**kwargs)
        self.morph_positions = morph_positions
        self.morph_colors = morph_colors

    def number_of_morphs(self) -> int:
        return max(len(self.morph_colors), len(self.morph_positions))

    def lower(self, object_data):
        ret = super(TriangularMeshGeometryWithMorphAttributes, self).lower(object_data=object_data)
        ret[u"data"][u"morphAttributes"] = {}
        if self.morph_positions is not None:
            ret[u"data"][u"morphAttributes"][u"position"] = [g.pack_numpy_array(pos.T) for pos in self.morph_positions]
        if self.morph_colors is not None:
            ret[u"data"][u"morphAttributes"][u"color"] = [g.pack_numpy_array(c.T) for c in self.morph_colors]
        return ret



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-14
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

import time
import numpy as np
from pathlib import Path
from example_robot_data.robots_loader import PandaLoader
from robomeshcat import Scene, Robot

"Meshcat can also create animations, that you can rewind forward/backward in the browser. This example shows how to "
"create such an animation, by changing the configuration of the robot."

scene = Scene()
robot = Robot(urdf_path=PandaLoader().df_path, mesh_folder_path=Path(PandaLoader().model_path).parent.parent)
scene.add_robot(robot)

with scene.animation(fps=1):  # start the animation with the current scene
    scene.render()

    robot[3] = np.deg2rad(-90)  # modify the scene
    scene.render()  # store the changes in the frame and start new frame modifications

    robot.opacity = 0.2
    scene.render()

    robot.opacity = 1.
    scene.render()

    robot.color = [0, 0, 0.7]
    scene.render()

    # then let's rotate camera around the robot for 10s;
    # note that setting camera will forbid user to rotate the scene
    for tt in np.linspace(0, 2 * np.pi, 10):
        scene.camera_pos[0] = np.sin(tt)
        scene.camera_pos[1] = np.cos(tt)
        scene.camera_pos[2] = 1
        scene.render()

    scene.reset_camera()  # resetting the camera will allow user interaction again

time.sleep(5)  # we need some time to send the animation to the browser



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-11
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#
from __future__ import annotations

import itertools
import time
from pathlib import Path
from tempfile import gettempdir

import imageio
import numpy as np
from PIL.Image import Image

import meshcat
from meshcat.animation import Animation, AnimationFrameVisualizer

from .object import Object, ArrayWithCallbackOnSetItem
from .robot import Robot


class Scene:
    def __init__(self, open: bool = True, wait_for_open: bool = True) -> None:
        super().__init__()
        self.objects: dict[str, Object] = {}
        self.robots: dict[str, Robot] = {}

        self.vis = meshcat.Visualizer()
        if open:
            self.vis.open()
        if wait_for_open:
            self.vis.wait()
        self.set_background_color()

        "Variables used to control the camera"
        self._camera_vis = self.vis["/Cameras/default"]
        self._camera_pose = ArrayWithCallbackOnSetItem(np.eye(4), cb=self._set_camera_transform)
        self._camera_pose_modified = False
        self._camera_zoom = 1.0

        " Variables used internally in case we are rendering to animation "
        self._animation: Animation | None = None
        self._animation_frame: AnimationFrameVisualizer | None = None
        self._animation_frame_counter: itertools.count | None = None

        " Variables used internally to write frames of the video "
        self._video_writer = None

    def add_object(self, obj: Object, verbose: bool = True):
        if verbose and obj.name in self.objects:
            print('Object with the same name is already inside the scene, it will be replaced. ')
        if verbose and self._animation is not None:
            print(
                'Adding objects while animating is not allowed.'
                'You need to add all objects before starting the animation.'
            )
            return
        self.objects[obj.name] = obj
        obj._set_vis(self.vis)
        obj._set_object()

    def remove_object(self, obj: Object, verbose: bool = True):
        if verbose and self._animation is not None:
            print('Removing object while animating is not allowed.')
            return
        obj._delete_object()
        self.objects.pop(obj.name)

    def add_robot(self, robot: Robot, verbose: bool = True):
        if verbose and robot.name in self.robots:
            print('Robot with the same name is already inside the scene, it will be replaced. ')
        self.robots[robot.name] = robot
        for obj in robot._objects.values():
            self.add_object(obj, verbose=verbose)

    def remove_robot(self, robot: Robot, verbose: bool = True):
        if verbose and self._animation is not None:
            print('Removing robots while animating is not allowed.')
            return
        self.robots.pop(robot.name)
        for obj in robot._objects.values():
            self.remove_object(obj, verbose=verbose)

    def render(self):
        """Render current scene either to browser, video or to the next frame of the animation."""
        if self._animation is not None:
            self._reset_all_properties()
            self._next_animation_frame()
        if self._video_writer is not None:
            self._video_writer.append_data(np.array(self.render_image()))

    #
    def video_recording(
        self, filename: Path | str | None = None, fps: int = 30, directory: Path | str | None = None, **kwargs
    ):
        """Return a context manager for video recording.
        The output filename is given by:
         1) filename parameter if it is not None
         2) directory/timestemp.mp4 if filename is None and directory is not None
         3) /tmp/timestemp if filename is None and directory is None
        """
        if filename is None:
            if directory is None:
                directory = gettempdir()
            filename = Path(directory).joinpath(time.strftime("%Y%m%d_%H%M%S.mp4"))
        return VideoContext(scene=self, fps=fps, filename=filename, **kwargs)

    def render_image(self) -> Image:
        return self.vis.get_image()

    def __getitem__(self, item):
        return self.objects[item] if item in self.objects else self.robots[item]

    def clear(self):
        """Remove all the objects/robots from the scene"""
        for r in list(self.robots.values()):
            self.remove_robot(r)
        for o in list(self.objects.values()):
            self.remove_object(o)

    def set_background_color(self, top_color=None, bottom_color=None):
        """Set background color of the visualizer. Use white background by default."""
        clr = Object._color_from_input(top_color, default=np.ones(3))
        self.vis["/Background"].set_property("top_color", clr.tolist())
        clr = Object._color_from_input(bottom_color, default=clr)
        self.vis["/Background"].set_property("bottom_color", clr.tolist())

    """=== The following set of functions handle animations ==="""

    def animation(self, fps: int = 30):
        """Return context of the animation that allow us to record animations.
        Usage:
            with scene.animation(fps=30):
                scene['obj'].pos = 3. # set properties of a first frame
                scene.render() # create a second frame of animation
                scene['obj'].pos = 3.
        """
        return AnimationContext(scene=self, fps=fps)

    def _next_animation_frame(self):
        """Close the current frame (if exists) and create a new one. Applicable only if animation exists. Set objects
        visualizer to the current frame."""
        assert self._animation is not None
        if self._animation_frame is not None:
            self._animation_frame.__exit__()
        self._animation_frame = self._animation.at_frame(self.vis, next(self._animation_frame_counter))
        self._animation_frame.__enter__()
        "Set all objects visualizers to the frame"
        for o in self.objects.values():
            o._set_vis(self._animation_frame)
        self._camera_vis = self._animation_frame["/Cameras/default"]

    def _close_animation(self):
        """Close the current frame and set objects visualizer to online."""
        assert self._animation is not None
        self._animation_frame.__exit__()
        self._animation_frame = None
        self._animation_frame_counter = None
        self._animation = None
        for o in self.objects.values():
            o._set_vis(self.vis)
        self._camera_vis = self.vis["/Cameras/default"]

    def _start_animation(self, fps):
        """Start animation instead of online changes."""
        self._animation_frame_counter = itertools.count()
        self._animation = Animation(default_framerate=fps)
        self._next_animation_frame()

    def _reset_all_properties(self):
        """Send all the properties to the browser. It is reset for all animation frames to ensure values do not change
        due to the interpolation by the properties set in the future."""
        for o in self.objects.values():
            o._reset_all_properties()
        self.camera_zoom = self._camera_zoom  # this will set the starting value of the property
        if self._camera_pose_modified:
            self.camera_pose = self._camera_pose

    """=== Following functions handle the camera control ==="""

    @property
    def camera_zoom(self):
        return self._camera_zoom

    @camera_zoom.setter
    def camera_zoom(self, v):
        self._camera_zoom = v
        self._set_property(self._camera_vis['rotated/<object>'], 'zoom', self._camera_zoom, 'number')

    @property
    def camera_pose(self):
        return self._camera_pose

    @camera_pose.setter
    def camera_pose(self, v):
        self._camera_pose[:, :] = v

    @property
    def camera_pos(self):
        return self._camera_pose[:3, 3]

    @camera_pos.setter
    def camera_pos(self, p):
        self._camera_pose[:3, 3] = p

    @property
    def camera_rot(self):
        return self._camera_pose[:3, :3]

    @camera_rot.setter
    def camera_rot(self, r):
        self._camera_pose[:3, :3] = r

    def reset_camera(self):
        """Reset camera to default pose and let user interact with it again."""
        self._camera_pose = ArrayWithCallbackOnSetItem(np.eye(4), cb=self._set_camera_transform)
        self._camera_vis.set_transform(self._camera_pose)
        self.camera_zoom = 1.0
        self._camera_enable_user_control()
        self._camera_pose_modified = False

    def _set_camera_transform(self):
        # disable human interaction if camera pose is set
        self._camera_vis.set_transform(self._camera_pose)
        self._camera_disable_user_control()
        self._camera_pose_modified = True

    def _camera_disable_user_control(self):
        self._set_property(self._camera_vis['rotated/<object>'], 'position', [0, 0, 0], 'vector')

    def _camera_enable_user_control(self):
        self._set_property(self._camera_vis['rotated/<object>'], 'position', [3, 1, 0], 'vector')

    @staticmethod
    def _set_property(element, key, value, prop_type='number'):
        """Set property of the object, handle animation frames set_property internally."""
        if isinstance(element, AnimationFrameVisualizer):
            element.set_property(key, prop_type, value)
        else:
            element.set_property(key, value)


class AnimationContext:
    """Used to provide 'with animation' capability for the viewer."""

    def __init__(self, scene: Scene, fps: int) -> None:
        super().__init__()
        self.scene: Scene = scene
        self.fps: int = fps

    def __enter__(self):
        self.scene._start_animation(self.fps)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Publish animation and clear all internal changes that were required to render to frame instead of online"""
        self.remove_clips_duplicates()
        self.scene.vis['animations/animation'].set_animation(self.scene._animation)
        self.scene._close_animation()

    def remove_clips_duplicates(self):
        """Meshcat doesn't like if same property as modified twice in the same frame - it does weird jumping in
        animation. In this function we find such a duplicates and keep only the last change of the property."""
        for clip in self.scene._animation.clips.values():
            for track in clip.tracks.values():
                indices_to_remove = set()
                last_f = None
                for i, f in enumerate(track.frames):
                    if f == last_f:
                        indices_to_remove.add(i - 1)
                    last_f = f
                if len(indices_to_remove) > 0:
                    track.frames = [f for i, f in enumerate(track.frames) if i not in indices_to_remove]
                    track.values = [v for i, v in enumerate(track.values) if i not in indices_to_remove]


class VideoContext:
    def __init__(self, scene: Scene, fps: int, filename: str | Path, **kwargs) -> None:
        super().__init__()
        self.scene = scene
        self.video_writer = imageio.get_writer(uri=filename, fps=fps, **kwargs)

    def __enter__(self):
        self.scene._video_writer = self.video_writer
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.video_writer.close()
        self.scene._video_writer = None



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-11
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

from __future__ import annotations

import itertools
from pathlib import Path
import trimesh
import numpy as np
from PIL import Image
import io
import meshcat.geometry as g
from meshcat.animation import AnimationFrameVisualizer


class Object:
    """Represent an object with arbitrary geometry that can be rendered in the meshcat."""

    id_iterator = itertools.count()

    def __init__(
        self,
        geometry=None,
        pose=None,
        color: list[float] | None = None,
        texture: g.ImageTexture | Path | None = None,
        opacity: float = 1.0,
        name: str | None = None,
    ) -> None:
        """Create an object with given geometry, pose, color and opacity.
        :param geometry
        :param pose - 4x4 pose of the object
        :param color - RGB color of the object either tuple of integers [0-255] or tuple of floats [0. - 1.]
        :param texture - alternative to color, you can specify texture that object uses to render as meshcat texture or
            as a path to png file.
        :param opacity - transparency of the object from 0. to 1.
        :param name - unique name of the object, it's created automatically if not specified
        """
        super().__init__()
        self.name = f'obj{next(self.id_iterator)}' if name is None else name
        self._vis = None  # either a visualization tree or the frame of the animation, is set by the scene

        "List of properties that could be updated for all objects"
        self._pose = ArrayWithCallbackOnSetItem(np.eye(4) if pose is None else pose, cb=self._set_transform)
        self._color = ArrayWithCallbackOnSetItem(self._color_from_input(color), cb=self._color_reset_on_set_item)
        self._opacity = opacity
        self._visible = True

        "Not updatable properties."
        self._texture = g.ImageTexture(g.PngImage.from_file(texture)) if isinstance(texture, (Path, str)) else texture
        self._geometry = geometry

    @staticmethod
    def _color_from_input(clr: list[float] | np.ndarray | None, default=None):
        """Modify input color to always be represented by numpy array with values from 0 to 1"""
        if clr is None:
            clr = np.random.uniform(0, 1, 3) if default is None else default
        clr = np.asarray(clr)
        assert clr.shape == (3,)
        if np.any(clr > 1) and clr.dtype == np.int:
            clr = clr.astype(dtype=np.float) / 255
        clr = clr.clip(0.0, 1.0)
        return clr

    def _set_vis(self, vis):
        """Set visualizer. Used internally to create frames of animation."""
        self._vis = vis[self.name]

    def _assert_vis(self):
        assert (
            self._vis is not None
        ), 'The properties of the object cannot be modified unless object is added to the scene.'

    def _set_object(self):
        """Create an object in meshcat and set all the initial properties."""
        self._assert_vis()
        self._vis.set_object(self._geometry, self._material)
        self._set_transform()

    def _delete_object(self):
        """Delete an object from meshcat."""
        self._vis.delete()

    def _set_transform(self):
        """Update transformation in the meshcat."""
        self._assert_vis()
        self._vis.set_transform(self._pose)

    def _set_property(self, key, value, prop_type='number', subpath='<object>'):
        """Set property of the object, handle animation frames set_property internally."""
        self._assert_vis()
        element = self._vis if subpath is None else self._vis[subpath]
        if self._is_animation():
            element.set_property(key, prop_type, value)
        else:
            element.set_property(key, value)

    def _is_animation(self):
        """Return true if rendering to the animation. Used internally to modify the way of assigning properties until
        PR https://github.com/rdeits/meshcat/pull/137 is merged."""
        return isinstance(self._vis, AnimationFrameVisualizer)

    @property
    def _material(self):
        if isinstance(self._geometry, g.Object):
            return None
        if self._texture is not None:
            return g.MeshLambertMaterial(map=self._texture, opacity=self.opacity)
        color = self.color.copy() * 255
        color = np.clip(color, 0, 255)
        return g.MeshLambertMaterial(
            color=int(color[0]) * 256**2 + int(color[1]) * 256 + int(color[2]), opacity=self.opacity
        )

    """=== Control of the pose ==="""

    @property
    def pose(self):
        return self._pose

    @pose.setter
    def pose(self, v):
        self._pose[:, :] = v

    @property
    def pos(self):
        return self._pose[:3, 3]

    @pos.setter
    def pos(self, p):
        self._pose[:3, 3] = p

    @property
    def rot(self):
        return self._pose[:3, :3]

    @rot.setter
    def rot(self, r):
        self._pose[:3, :3] = r

    """=== Control of the object visibility ==="""

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, v):
        self._visible = bool(v)
        self._set_property('visible', value=self._visible, prop_type='boolean')

    def hide(self):
        self.visible = False

    def show(self):
        self.visible = True

    """=== Control of the object transparency ==="""

    @property
    def opacity(self):
        return self._opacity

    @opacity.setter
    def opacity(self, v):
        self._opacity = v
        if self._is_animation():
            self._set_property('material.opacity', value=self._opacity, prop_type='number')
        else:
            self._set_object()  # only way how to update opacity online is to reset the object

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, v):
        self._color[:] = self._color_from_input(v)

    def _color_reset_on_set_item(self):
        if self._is_animation():
            self._set_property('material.color', value=self.color.tolist(), prop_type='vector')
        else:
            self._set_object()  # only way how to update color online is to reset the object

    def _reset_all_properties(self):
        """Reset all properties in the meshcat, i.e. call setters with the same variable. Useful for the animation
        frames."""
        self.pose = self.pose
        self.color = self.color
        self.opacity = self.opacity
        self.visible = self.visible

    """=== Helper functions to create basic primitives ==="""

    @classmethod
    def create_cuboid(
        cls,
        lengths: list[float] | float,
        pose=None,
        color: list[float] | None = None,
        texture: g.ImageTexture | Path | None = None,
        opacity: float = 1.0,
        name: str | None = None,
    ):
        """Create cuboid with a given size."""
        box = g.Box(lengths=[lengths] * 3 if isinstance(lengths, (float, int)) else lengths)
        return cls(box, pose=pose, color=color, texture=texture, opacity=opacity, name=name)

    @classmethod
    def create_sphere(
        cls,
        radius: float,
        pose=None,
        color: list[float] | None = None,
        texture: g.ImageTexture | Path | None = None,
        opacity: float = 1.0,
        name: str | None = None,
    ):
        """Create a sphere with a given radius."""
        return cls(g.Sphere(radius=radius), pose=pose, color=color, texture=texture, opacity=opacity, name=name)

    @classmethod
    def create_cylinder(
        cls,
        radius: float,
        length: float,
        pose=None,
        color: list[float] | None = None,
        texture: g.ImageTexture | Path | None = None,
        opacity: float = 1.0,
        name: str | None = None,
    ):
        """Create a cylinder with a given radius and length. The axis of rotational symmetry is aligned with the z-axis
        that is common in robotics. To achieve that, we create a mesh of a cylinder instead of using meshcat cylinder
        that is aligned with y-axis."""
        mesh: trimesh.Trimesh = trimesh.creation.cylinder(radius=radius, height=length, sections=50)
        exp_obj = trimesh.exchange.obj.export_obj(mesh)
        mesh = g.ObjMeshGeometry.from_stream(trimesh.util.wrap_as_stream(exp_obj))
        return cls(mesh, pose=pose, color=color, texture=texture, opacity=opacity, name=name)

    @classmethod
    def create_mesh(
        cls,
        path_to_mesh: str | Path,
        scale: float | list[float] = 1.0,
        pose=None,
        color: list[float] | None = None,
        texture: g.ImageTexture | Path | None = None,
        opacity: float = 1.0,
        name: str | None = None,
    ):
        """Create a mesh object by loading it from the :param path_to_mesh. Loading is performed by 'trimesh' library
        internally."""
        try:
            mesh: trimesh.Trimesh = trimesh.load(path_to_mesh, force='mesh')
        except ValueError as e:
            if str(e) == 'File type: dae not supported':
                print(
                    'To load DAE meshes you need to install pycollada package via '
                    '`conda install -c conda-forge pycollada`'
                    ' or `pip install pycollada`'
                )
            raise
        except Exception as e:
            print(
                f'Loading of a mesh failed with message: \'{e}\'. '
                f'Trying to load with with \'ignore_broken\' but consider to fix the mesh located here:'
                f' \'{path_to_mesh}\'.'
            )
            mesh: trimesh.Trimesh = trimesh.load(path_to_mesh, force='mesh', ignore_broken=True)

        mesh.apply_scale(scale)

        try:  # try to get texture from the mesh
            if hasattr(mesh, 'visual') and hasattr(mesh.visual, 'material') and texture is None:
                n = 100

                def xy_to_uv(xy, w=n, h=n):
                    u = xy[..., 0] / (w - 1)
                    v = 1 - xy[..., 1] / (h - 1)
                    return np.stack((u, v), axis=-1)

                X, Y = np.meshgrid(np.arange(n), np.arange(n))
                uv = xy_to_uv(np.stack((X, Y), axis=-1))
                data = mesh.visual.material.to_color(uv.reshape(-1, 2))
                if data.shape == (4,):
                    data = np.tile(data, n * n).reshape(n, n, 4)
                else:
                    data = data.reshape(n, n, 4)
                b = io.BytesIO()
                Image.fromarray(data).save(b, 'png')
                texture = g.ImageTexture(g.PngImage(b.getvalue()))
        except Exception:
            pass

        try:
            exp_obj = trimesh.exchange.obj.export_obj(mesh)
        except ValueError:
            exp_obj = trimesh.exchange.obj.export_obj(mesh, include_texture=False)

        return cls(
            g.ObjMeshGeometry.from_stream(trimesh.util.wrap_as_stream(exp_obj)),
            pose=pose,
            color=color,
            texture=texture,
            opacity=opacity,
            name=name,
        )


class ArrayWithCallbackOnSetItem(np.ndarray):
    def __new__(cls, input_array, cb=None):
        obj = np.asarray(input_array).view(cls)
        obj.cb = cb
        return obj

    def __setitem__(self, *args, **kwargs):
        super().__setitem__(*args, **kwargs)
        self.cb()

    def __array_finalize__(self, obj, **kwargs):
        if obj is None:
            return
        self.cb = getattr(obj, 'cb', None)
